#include "feature.h"

//-------------------------------------------------------------------------------------------------
void AudiPeriph(float *input, int sigLength, gammaTone *fChan, filter bPass, feature *corrLgm)
{	
	float lowerERB, upperERB, spaceERB, cf, phon;
	int chan;
	float *gOut, *hc, *hev;

	gOut = new float[sigLength];
	hc = new float[sigLength];
	hev = new float[sigLength];

	// Calculate the parameters of gammatone filters
	middleEar loudFunc = initMiddleEar();

	lowerERB = HzToERBRate(MINCF);
	upperERB = HzToERBRate(MAXCF);
  
	spaceERB = (NUMBER_CHANNEL > 1) ? (upperERB-lowerERB)/(NUMBER_CHANNEL-1) : 0;
	
	for(chan=0; chan<NUMBER_CHANNEL; chan++)
	{
		cf = ERBRateToHz(lowerERB + chan*spaceERB);
		fChan[chan].cf = cf;
		fChan[chan].bw = 24.7*(cf*0.00437 + 1.0) * BW_CORRECTION;
		fChan[chan].delay = 1.5/PI/fChan[chan].bw*float(SAMPLING_FREQUENCY);
		
		phon = (loudnessLevelInPhons(cf, loudFunc) - DB);
		fChan[chan].midEarCoeff = pow(10, phon/20); 
	}

	// Bandpass filtering, auditory nerve transduction, and feature extraction
	for(chan=0; chan<NUMBER_CHANNEL; chan++)
	{
		if ((chan%8)==0) printf(".");

		gammaToneFilter(input, sigLength, fChan[chan], gOut);
		hairCell(gOut, sigLength, hc);
		envelope(hc, sigLength, bPass, hev);

		computeFeature(gOut, hc, hev, sigLength, WINDOW, chan, corrLgm);
	}

	int numFrame = sigLength/OFFSET;

	crossCorr(corrLgm, numFrame);

	delete gOut;
	delete hc;
	delete hev;
}

middleEar initMiddleEar()   
{
	// parameters of equal-loudness functions from BS3383,"Normal equal-loudness level
	// contours for pure tones under free-field listening conditions", table 1.
	middleEar lF;
	lF.f[0]=20.0;     lF.af[0]=2.347;  lF.bf[0]=0.00561;   lF.tf[0]=74.3;
	lF.f[1]=25.0;     lF.af[1]=2.190;  lF.bf[1]=0.00527;   lF.tf[1]=65.0;
	lF.f[2]=31.5;     lF.af[2]=2.050;  lF.bf[2]=0.00481;   lF.tf[2]=56.3;
	lF.f[3]=40.0;     lF.af[3]=1.879;  lF.bf[3]=0.00404;   lF.tf[3]=48.4;
	lF.f[4]=50.0;     lF.af[4]=1.724;  lF.bf[4]=0.00383;   lF.tf[4]=41.7;
	lF.f[5]=63.0;     lF.af[5]=1.579;  lF.bf[5]=0.00286;   lF.tf[5]=35.5;
	lF.f[6]=80.0;     lF.af[6]=1.512;  lF.bf[6]=0.00259;   lF.tf[6]=29.8;
	lF.f[7]=100.0;    lF.af[7]=1.466;  lF.bf[7]=0.00257;   lF.tf[7]=25.1;
	lF.f[8]=125.0;    lF.af[8]=1.426;  lF.bf[8]=0.00256;   lF.tf[8]=20.7;
	lF.f[9]=160.0;    lF.af[9]=1.394;  lF.bf[9]=0.00255;   lF.tf[9]=16.8;
	lF.f[10]=200.0;   lF.af[10]=1.372; lF.bf[10]=0.00254;  lF.tf[10]=13.8;
	lF.f[11]=250.0;   lF.af[11]=1.344; lF.bf[11]=0.00248;  lF.tf[11]=11.2;
	lF.f[12]=315.0;   lF.af[12]=1.304; lF.bf[12]=0.00229;  lF.tf[12]=8.9;
	lF.f[13]=400.0;   lF.af[13]=1.256; lF.bf[13]=0.00201;  lF.tf[13]=7.2;
	lF.f[14]=500.0;   lF.af[14]=1.203; lF.bf[14]=0.00162;  lF.tf[14]=6.0;
	lF.f[15]=630.0;   lF.af[15]=1.135; lF.bf[15]=0.00111;  lF.tf[15]=5.0;
	lF.f[16]=800.0;   lF.af[16]=1.062; lF.bf[16]=0.00052;  lF.tf[16]=4.4;
	lF.f[17]=1000.0;  lF.af[17]=1.000; lF.bf[17]=0.00000;  lF.tf[17]=4.2;
	lF.f[18]=1250.0;  lF.af[18]=0.967; lF.bf[18]=-0.00039; lF.tf[18]=3.7;
	lF.f[19]=1600.0;  lF.af[19]=0.943; lF.bf[19]=-0.00067; lF.tf[19]=2.6; 
	lF.f[20]=2000.0;  lF.af[20]=0.932; lF.bf[20]=-0.00092; lF.tf[20]=1.0;
	lF.f[21]=2500.0;  lF.af[21]=0.933; lF.bf[21]=-0.00105; lF.tf[21]=-1.2; 
	lF.f[22]=3150.0;  lF.af[22]=0.937; lF.bf[22]=-0.00104; lF.tf[22]=-3.6;
	lF.f[23]=4000.0;  lF.af[23]=0.952; lF.bf[23]=-0.00088; lF.tf[23]=-3.9; 
	lF.f[24]=5000.0;  lF.af[24]=0.974; lF.bf[24]=-0.00055; lF.tf[24]=-1.1;
	lF.f[25]=6300.0;  lF.af[25]=1.027; lF.bf[25]=0.00000;  lF.tf[25]=6.6;
	lF.f[26]=8000.0;  lF.af[26]=1.135; lF.bf[26]=0.00089;  lF.tf[26]=15.3;
	lF.f[27]=10000.0; lF.af[27]=1.266; lF.bf[27]=0.00211;  lF.tf[27]=16.4;
	lF.f[28]=12500.0; lF.af[28]=1.501; lF.bf[28]=0.00488;  lF.tf[28]=11.6;
	return(lF);
}

float loudnessLevelInPhons(float freq, middleEar lF)
{
	//Uses linear interpolation of the look-up tables to compute the loudness level, in phons, 
	//of a pure tone of frequency freq using the reference curve for sound pressure level dB.
	//The equation is taken from section 4 of BS3383.

	int i;
	float afy, bfy, tfy;
	float ratio;

	if ((freq<20.0) | (freq>12500.0)) {
		fprintf(stderr,"Can't compute a outer/middle ear gain for that frequency\n");
		exit(0);
	}
	
	i = 0;
	while (lF.f[i] < freq) i++;

	ratio = (freq - lF.f[i-1]) / (lF.f[i] - lF.f[i-1]);
	afy = lF.af[i-1] + ratio * (lF.af[i] - lF.af[i-1]);
	bfy	= lF.bf[i-1] + ratio * (lF.bf[i] - lF.bf[i-1]);
	tfy	= lF.tf[i-1] + ratio * (lF.tf[i] - lF.tf[i-1]);
	
	return(4.2 + afy*(DB - tfy) / (1.0 + bfy*(DB - tfy)) );
}

float HzToERBRate(float Hz)
{
	return( 21.4*log10(Hz*0.00437 + 1.0) );
}

float ERBRateToHz(float ERBRate)
{
	return( (pow(10, ERBRate/21.4) - 1) / 0.00437 );
}

void gammaToneFilter(float *input, int sigLength, gammaTone fChan, float *output)
{
	float dt, twoPiT, gain, z;
	float f1, f2, x[4], y[4];
	int i, n; 
	int delay=int(fChan.delay);
	
	dt= 1/float(SAMPLING_FREQUENCY);

	twoPiT=2 * PI * dt;
	gain = fChan.midEarCoeff * pow(twoPiT * fChan.bw, 4) / 3.0;
	z = exp(-twoPiT * fChan.bw);

	f1 = cos(fChan.cf * twoPiT) * z;
	f2 = sin(fChan.cf * twoPiT) * z;

	for (i=0; i<4; i++)
	{
		fChan.p[i] = 0;
		fChan.q[i] = 0;
	}

	for (n=0; n<(sigLength+delay); n++)
	{
		if (n>=delay) output[n-delay] = fChan.p[3] * gain;
		for (i=0; i<4; i++)
		{
			x[i] = f1*fChan.p[i] - f2*fChan.q[i];
			y[i] = f2*fChan.p[i] + f1*fChan.q[i];
		}

		if (n<sigLength)
		{
			fChan.p[0] = input[n] * f1 + x[0];
			fChan.q[0] = input[n] * f2 + y[0];
		}

		else
		{
			fChan.p[0] = x[0];
			fChan.q[0] = y[0];
		}
		
		fChan.p[1] = fChan.p[0] + x[1];
		fChan.q[1] = fChan.q[0] + y[1];
		
		fChan.p[2] = fChan.p[1] + x[1] + x[2];
		fChan.q[2] = fChan.q[1] + y[1] + y[2];
		
		fChan.p[3] = fChan.p[2] + x[1] + 2*x[2] + x[3];
		fChan.q[3] = fChan.q[2] + y[1] + 2*y[2] + y[3];
	}
}

void hairCell(float *input, int sigLength, float *output)
{
	float dt, ymdt, xdt, ydt, lplusrdt, rdt, gdt, hdt;
	float kt, c, q, w;
	float replenish, eject, reuptakeandloss, reuptake, reprocess;
	int n;

	dt = 1/float(SAMPLING_FREQUENCY);
	ymdt = MED_Y * MED_M * dt;
	xdt = MED_X * dt;
	ydt = MED_Y * dt;
	lplusrdt = (MED_L + MED_R) * dt;
	rdt = MED_R * dt;
	gdt = MED_G * dt;
	hdt = MED_H; // should be multiplied by dt really

	//initialize variables
	kt = MED_G * MED_A / (MED_A + MED_B);
	c = MED_M * MED_Y * kt / (MED_L * kt + MED_Y * (MED_L + MED_R ));
	q = c * (MED_L + MED_R)/kt;
	w = c * MED_R / MED_X;

	// calculate output
	for (n=0; n<sigLength; n++)
	{
		kt = ((input[n]+MED_A) > 0.0) ? (gdt * (input[n]+MED_A) / (input[n]+MED_A+MED_B)) : 0;

		replenish = (q<MED_M) ? (ymdt - ydt*q) : 0;
        
		eject = kt * q;
		reuptakeandloss = lplusrdt * c;
		reuptake = rdt * c;
		reprocess = xdt * w;

		q = q + replenish - eject + reprocess;
		if (q < 0.0) q = 0.0;

		c = c + eject - reuptakeandloss;
		if (c < 0.0) c = 0.0;
   
        w = w + reuptake - reprocess;
		if (w < 0.0) w = 0.0;
    
		output[n] = hdt * c;
	}
}

//------------------------------------------------------------------------------------------------
void envelope(float *input, int sigLength, filter bPass, float *output)
{
	// number of batches for envelope extraction
	int tim = ceil(float(sigLength)/float(bPass.nFFT - bPass.fLength));

	for(int t=0; t<tim; t++)
	{
		// half-wave rectify
		for(int n=0; n<bPass.nFFT; n++)
		{
			int m=t*(bPass.nFFT - bPass.fLength)+n;

			bPass.inputR[n] = ( (m < (sigLength-1)) && (m>0) &&(input[m]>0) ) ? input[m]: 0;
			bPass.inputI[n] = 0;
		}
		
		// banspass filtering using FFT
		fft(bPass.inputR, bPass.inputI, bPass.nOrder, 1);

		for(n=0; n<bPass.nFFT; n++)
		{
			float f1 = bPass.rValue[n] * bPass.inputR[n] - bPass.iValue[n] * bPass.inputI[n];
			float f2 = bPass.rValue[n] * bPass.inputI[n] + bPass.iValue[n] * bPass.inputR[n];

			bPass.inputR[n] = f1;
			bPass.inputI[n] = f2;
		}
		
		fft(bPass.inputR, bPass.inputI, bPass.nOrder, -1);

		for(n=0; n<(bPass.nFFT - bPass.fLength/2); n++)
		{
			int m = t * (bPass.nFFT - bPass.fLength) + n;
			if(n<(bPass.fLength/2))
			{
				if (t==0) output[m] = bPass.inputR[n+bPass.fLength/2]/float(bPass.nFFT);
			}
			else if (m<sigLength) output[m] = bPass.inputR[n+bPass.fLength/2]/float(bPass.nFFT);
		}
	}
}

void computeFeature(float *gOut, float *hc, float *hev, int sigLength, int winsize, int chan, feature *corrLgm)
{
	int numFrame, frame;
	int delay, step, tim;
	float acf[NFFT_ACF], eng[MAX_DELAY];

	numFrame = sigLength/OFFSET;

	for(frame=0; frame<numFrame; frame++)
	{
		// Calculate energy
		corrLgm[frame].eng[chan]=0;
		for(step=-OFFSET; step<OFFSET; step++)
		{
			int tim=frame*OFFSET+step;
			if ( (tim>=0) && (tim<sigLength))
			{
				corrLgm[frame].eng[chan] += gOut[tim]*gOut[tim];
			}
		}

		// Calculate autocorrelation using FFT
		fftACF(hc, sigLength, winsize, frame, acf);

		for(delay=0; delay<MAX_DELAY; delay++)
			corrLgm[frame].acf[chan][delay] = acf[delay]/acf[0];

		fftACF(hev, sigLength, winsize, frame, acf);

		for(delay=0; delay<MAX_DELAY; delay++)
			corrLgm[frame].acfEv[chan][delay] = acf[delay]/acf[0];
	}
}

void fftACF(float *sig, int sigLength, int winsize, int frame, float *inputR)
{
	int step, tim;
	float inputI[NFFT_ACF], inputSR[NFFT_ACF], inputSI[NFFT_ACF];

	for(step=0; step<NFFT_ACF; step++)
	{
		inputR[step]=0; inputSR[step]=0;
		inputI[step]=0; inputSI[step]=0;
	}

	for(step=0; step<(MAX_DELAY+winsize); step++)
	{
		tim = (frame+2) * OFFSET - (step+1);
		if ( (tim>=0) && (tim<sigLength) && (step<winsize) ) inputR[step]=sig[tim];
		if ((tim>=0) && (tim<sigLength) ) inputSR[step]=sig[tim];
	}

	fft(inputR, inputI, ORDERFFT_ACF, 1);
	fft(inputSR, inputSI, ORDERFFT_ACF, 1);

	for(step=0; step<NFFT_ACF; step++)
	{
		float f1 = inputR[step]*inputSR[step] + inputI[step]*inputSI[step];
		float f2 = inputR[step]*inputSI[step] - inputI[step]*inputSR[step];

		inputR[step]=f1;
		inputI[step]=f2;
	}

	fft(inputR, inputI, ORDERFFT_ACF, -1);
}

void crossCorr(feature *corrLgm, int numFrame)
{
	int chan, frame, delay;
	float acfHc[NUMBER_CHANNEL][MAX_DELAY], acfEv[NUMBER_CHANNEL][MAX_DELAY], sumHc,sumEv;

	for(chan=0; chan<NUMBER_CHANNEL; chan++)
		for(frame=0; frame<5; frame++)
		{
			corrLgm[frame].cross[chan] = corrLgm[frame].crossEv[chan] = 0;
		}
		
	for(frame=1; frame<numFrame; frame++)
	{
		// Normalize autocorrelation
		for(chan=0; chan<NUMBER_CHANNEL; chan++)
		{
			sumHc = 0;
			sumEv = 0;
			for(delay=0; delay<MAX_DELAY; delay++) 
			{
				sumHc += corrLgm[frame].acf[chan][delay];
				sumEv += corrLgm[frame].acfEv[chan][delay];
			}
				
			sumHc /= float(MAX_DELAY);
			sumEv /= float(MAX_DELAY);

			for(delay=0; delay<MAX_DELAY; delay++) 
			{
				acfHc[chan][delay] = corrLgm[frame].acf[chan][delay] - sumHc;
				acfEv[chan][delay] = corrLgm[frame].acfEv[chan][delay] - sumEv;
			}
				
			sumHc = 0;
			sumEv = 0;
			for(delay=0; delay<MAX_DELAY; delay++)
			{
				sumHc += acfHc[chan][delay] * acfHc[chan][delay];
				sumEv += acfEv[chan][delay] * acfEv[chan][delay];
			}
			
			sumHc = sqrt(sumHc / float(MAX_DELAY));
			sumEv = sqrt(sumEv / float(MAX_DELAY));
			
			if(sumHc !=0 )
			{
				for(delay=0; delay<MAX_DELAY; delay++)
					acfHc[chan][delay] /= sumHc;
			}
			if(sumEv !=0 )
			{
				for(delay=0; delay<MAX_DELAY; delay++)
					acfEv[chan][delay] /= sumEv;
			}
		}
	
		// Calculate cross-channel correlation
		for(chan=0; chan<(NUMBER_CHANNEL-1); chan++)
		{
			corrLgm[frame].cross[chan] = 0;
			corrLgm[frame].crossEv[chan] = 0;

			for(delay=0; delay<MAX_DELAY; delay++)
			{
				corrLgm[frame].cross[chan] += acfHc[chan][delay] * acfHc[chan+1][delay];
				corrLgm[frame].crossEv[chan] += acfEv[chan][delay] * acfEv[chan+1][delay];
			}
		
			corrLgm[frame].cross[chan] /= float(MAX_DELAY);
			corrLgm[frame].crossEv[chan] /= float(MAX_DELAY);
		}
		
		corrLgm[frame].cross[NUMBER_CHANNEL-1] = 0;;
		corrLgm[frame].crossEv[NUMBER_CHANNEL-1] = 0;;
	}
}

//------------------------------------------------------------------------------------------------
void resynthesis(float *input, int sigLength, gammaTone *fChan, feature *corrLgm, float *output)
{
	int n, chan, frame;
	float *temp, *weight, *reverse;

	int numFrame=sigLength/OFFSET;

	temp = new float[sigLength];
	weight = new float[sigLength];
	reverse = new float[sigLength];

	for(n=0; n<sigLength; n++)
		output[n]=0;

	for(chan=0; chan<NUMBER_CHANNEL; chan++)
	{
		// gammatone filtering
		gammaToneFilter(input, sigLength, fChan[chan], temp);

		// time reverse
		for(n=0; n<sigLength; n++)
			reverse[sigLength - n - 1] = temp[n] / fChan[chan].midEarCoeff;
        
		// gammatone filtering for the second time
		gammaToneFilter(reverse, sigLength, fChan[chan], temp);

		// time reverse for the second time
		for(n=0; n<sigLength; n++)
			reverse[sigLength - n - 1] = temp[n] / fChan[chan].midEarCoeff;
        
		// calculate weight for each sample of phase-corrected response according to the mask
		for(n=0; n<sigLength; n++) 
			weight[n] = 0.0;
        
		for(frame=0; frame<numFrame; frame++)
		{
			if(corrLgm[frame].stream[chan]>0)
			{
				if(frame > 0){
        			for(n=0; n<OFFSET; n++)
						weight[(frame-1)*OFFSET + n] += 0.5 * (1.0 + cos(n * PI/(OFFSET) + PI));
				}
				
				for(n=OFFSET; n<WINDOW; n++)
					weight[(frame-1)*OFFSET + n] += 0.5 * (1.0 + cos((n-OFFSET) * PI/(OFFSET)));
			}
		}

		// adding weighted response together
		for(n=0; n<sigLength; n++)
			output[n] += weight[n] * reverse[n];
	}
	
	delete temp;
	delete weight;
	delete reverse;
}
